/* 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jPong;

import java.io.IOException;

import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.GLEventListener;
import javax.media.opengl.glu.GLU;
import javax.vecmath.Vector3f;

/**
 * The main camera class, that allows mouse orbiting,
 * zooming and panning within the 3D environment.<br>
 * 
 * This viewer allows orbiting around the center of the camera view.
 * It will also display the Arrows for the center of the view.
 * <br>
 * This version is Z-up oriented. (as in not Y-up oriented)
 * <br>
 * Tested with jogl-1.1.1
 * <br><br>
 * This viewer controls just the JPong Players
 * 
 * @author TomDalek, Copyright (C) 2010.
 * @see <a href="http://www.gnu.org/licenses/gpl.html">The GNU Public License (GPL)</a>
 * 
 */
public class Viewer  implements GLEventListener {
	GLAutoDrawable drawable;	
	
	Player controledPlayer = null;
	boolean move = false;
	
	boolean flatShading = true;
	boolean cullFaces = true;
	boolean wireframe = false;
	boolean average = false;
	
	/**
	 * The location of the camera is stored as: <br>
	 * 1) Location of the center of the view
	 * as a Cartesian translation vector (trsX, trsY, trsZ). <br>
	 * 2) Location of the camera relative to this center
	 * as a spherical point (rotX, rotZ, zoom).
	 * <br><br>
	 * This seamed more intuitive approach for an orbiting viewer,
	 * then the glu's LookAt() which uses both points (camera and center)
	 * as global (absolute) points in the universe.
	 * 
	 * @author TomDalek
	 */
	class Location{
		float trsX = 0.0f;
		float trsY = 0.0f;
		float trsZ = 0.0f;
		float rotX = 0.0f;
		float rotZ = 0.0f;
		float zoom = 0.0f;
		
		Location(){			
		}
		
		Location(float trsX, float trsY, float trsZ,
				 float rotX, float rotZ, float zoom){
			this.trsX = trsX;
			this.trsY = trsY;
			this.trsZ = trsZ;
			this.rotX = rotX;
			this.rotZ = rotZ;
			this.zoom = zoom;
		}
		
		Location(Location location){
			this.trsX = location.trsX;
			this.trsY = location.trsY;
			this.trsZ = location.trsZ;
			this.rotX = location.rotX;			
			this.rotZ = location.rotZ;
			this.zoom = location.zoom;
		}
		
		void setTo(Location location){
			this.trsX = location.trsX;
			this.trsY = location.trsY;
			this.trsZ = location.trsZ;
			this.rotX = location.rotX;
			this.rotZ = location.rotZ;
			this.zoom = location.zoom;
		}
	}
	
	// Home camera position
	Location home = new Location(   0.0f, 15.0f, 20.0f, -45.0f, 0.0f,  -180.0f);
	
	
	// Present camera position (center)
	Location cam = new Location(home);
	Location camMin = new Location(  -200.0f, -200.0f, -200.0f,	-180.0f,   0.0f, -600.0f);
	Location camMax = new Location(   200.0f,  200.0f,  200.0f,    0.0f, 359.0f,   -2.0f);
	
	Location front = new Location(	0.0f,    0.0f, 0.0f, -90.0f,   0.0f, home.zoom);	
	Location top =   new Location(	0.0f,    0.0f, 0.0f,   0.0f,   0.0f, home.zoom);
	Location right = new Location(  0.0f,    0.0f, 0.0f, -90.0f, 270.0f, home.zoom);
	
	// Zoom in, zoom out step
	private float zoomStep = 1.2f;
	
	// Mouse state
	private int mx, my;
	private boolean mLeftPressed = false;
	private boolean mMiddlePressed = false;
	private boolean mRightPressed = false;


	
	// Window state	
	public int winWidth     = 800;
	public int winHeight    = 500;
	float winAspect  = (float)winWidth / (float)winHeight;
	
	// Camera state
	// GLvoid gluPerspective( GLdouble fovy, GLdouble aspect, GLdouble nearClip, GLdouble farClip )
	float fovy     = 60.0f;
	float nearClip = 1.0f;
	float farClip  = 2000.0f;
	
	/**
	 * The height of the view at the center of gluPerspective cone.
	 */
	float viewHeight = (float) (2.0 * Math.abs(cam.zoom) * Math.tan(Math.toRadians(fovy/2.0f)));
	
	/**
	 * Used to correlate mouse movement in pixels with viewers translation in GL coordinates.
	 * <br><br>
	 * I want to be able to pan the view as if I'm grabbing the plane of the view.
	 * But because I'm using gluPerspective I need to know the ratio between
	 * what I see and what I get. :) 
	 * <br>
	 * As in I see some portion of a scene (view),
	 * and what I get is mouse movement in pixels (window). 
	 */
	float viewToWinRatio = (float)viewHeight / (float)winHeight;

	
	Viewer(){		
		//System.err.println("winAspect is: " + winAspect);
		//System.err.println("viewHeight is: " + viewHeight);
		//System.err.println("viewToWinRatio is: " + viewToWinRatio);	
	}
	
	/**
	 * Recalculate winAspect, viewHeight and viewToWinRatio.
	 */
	void refreshViewToWinRatio(){
		if (winHeight <= 0) { // avoid a divide by zero error!
			winHeight = 1;
		}		
		winAspect  = (float)winWidth / (float)winHeight;
		viewHeight = (float) (2.0 * Math.abs(cam.zoom) * Math.tan(Math.toRadians(fovy/2.0f)));
		viewToWinRatio = (float)viewHeight / (float)winHeight;
	}

	public void setLeftPressed(boolean mLeftPressed, int x, int y) {
		mx = x;
		my = y;
		this.mLeftPressed = mLeftPressed;
		//System.err.println("LeftPressed" + " x=" + viewer.mx + " y=" + viewer.my);	
		
	
		float px =  x * 0.395f - 100.0f;
		float py = -y * 0.4f + 40.0f;
		
		controledPlayer = MasterListener.myWorldRenderer.myWorld.player1;
		
		if (mLeftPressed) {
			try {
				MasterListener.out.writeObject("add_joint");
				MasterListener.out.writeFloat(px);
				MasterListener.out.writeFloat(py);
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
			//controledPlayer.addJoint(px, py);
		}
		else {
			try {
				MasterListener.out.writeObject("remove_joint");
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			//controledPlayer.removeJoint();
		}
	}
	
	public void setMiddlePressed(boolean mMiddlePressed, int x, int y) {
		mx = x;
		my = y;
		this.mMiddlePressed = mMiddlePressed;
	}
	

	public void setRightPressed(boolean mRightPressed, int x, int y) {
		mx = x;
		my = y;
		this.mRightPressed = mRightPressed;

	}

	public boolean getmLeftPressed() {
		return mLeftPressed;
	}

	public boolean getmMiddlePressed() {
		return mMiddlePressed;
	}

	public boolean getmRightPressed() {
		return mRightPressed;
	}
	
	/**
	 * @return A normal vector pointing towards the camera.
	 */
	MyVector3f getVectorToMe(){
		MyVector3f vector = new MyVector3f();
		vector.setFromPolor(cam.rotZ, cam.rotX, 1.0f);
		return vector;
	}
	
	/**
	 *  Used for panning the view along the local plane of view.
	 * @return A normal vector pointing up within the local viewers plane.
	 */
	MyVector3f getVectorLocalUp(){
		MyVector3f vector = new MyVector3f();
		vector.setFromPolor(cam.rotZ+180.0f, -90.0f-cam.rotX, 1.0f);
		return vector;
	}
	
	/**
	 * Used for panning the view along the local plane of view.
	 * @return A normal vector pointing right within the local viewers plane.
	 */
	MyVector3f getVectorLocalRight(){
		MyVector3f vector = new MyVector3f();
		
		vector.cross(getVectorLocalUp(), getVectorToMe());
		vector.normalize();
		
		return vector;
	}
	
	

	public void panView(int x, int y) {
		int dx = mx - x;
		int dy = my - y;
		
		//===========================
		// translation XY within cameras local plane
		float localTrsX = dx * viewToWinRatio;	// Translate X 
		float localTrsY = dy * viewToWinRatio;	// Translate Y
		
		// global displacement of the cameras center along local X
		MyVector3f displaceX = getVectorLocalRight();
		displaceX.scale(localTrsX);
		// shift the orientation of mouse x movement (so that left is left)
		displaceX.scale(-1.0f);		
		
		// global displacement of the cameras center along local Y
		MyVector3f displaceY = getVectorLocalUp();
		displaceY.scale(localTrsY);
		
		// global displacement of the cameras center on the local XY plane
		MyVector3f displaceXY = new MyVector3f();				
		displaceXY.add(displaceX, displaceY);
		//===========================

		cam.trsX += displaceXY.x;
		cam.trsY += displaceXY.y;
		cam.trsZ += displaceXY.z;		
		
		mx = x;
		my = y;
		
		drawable.display();
	}
	
	/**
	 * This moves the the center of the view in and out along the cameras local Z axis.
	 * (As in panning is local XY translation, then flyIn is local Z translation.)
	 */
	public void flyInView(int x, int y){
		int dy = my - y;
		
		float localTrsZ = dy * viewToWinRatio;	// Translate X 
		
		MyVector3f displaceZ = getVectorToMe();
		displaceZ.scale(-localTrsZ);
		
		// Z is up
		cam.trsX += displaceZ.x;
		cam.trsY += displaceZ.y;
		cam.trsZ += displaceZ.z;	
		
		my = y;
		
		drawable.display();
	}

	public void rotateView(int x, int y) {
		int dx = mx - x;
		int dy = my - y;
		
		cam.rotX -= dy; // Mouse Y-axis is like grabbing front of the model.
		if(cam.rotX > camMax.rotX)
			cam.rotX = camMax.rotX;
		if(cam.rotX < camMin.rotX)
			cam.rotX = camMin.rotX;
		
		cam.rotZ -= dx; // Left is left. Right is right.
		if(cam.rotZ > camMax.rotZ)
			cam.rotZ = camMin.rotZ;
		if(cam.rotZ < camMin.rotZ)
			cam.rotZ = camMax.rotZ;
		
		mx = x;
		my = y;
		
		//System.err.println("mx=" + mx + "    my=" + my);
		//System.err.println("rotX=" + rotX + "    rotZ=" + rotZ);
		
		drawable.display();
	}
	
	/**
	 * Controls the distance between camera and cameras center point.
	 * @param wheelRotation
	 */
	public void rollZoomView(int wheelRotation) {        
	    //System.err.println(notches);
	    
		/* jumpy
	    if (wheelRotation < 0) //Roll up (zoom in)
			cam.zoom /= zoomStep;
		if (wheelRotation > 0) //Roll down (zoom out)
			cam.zoom *= zoomStep;
		
		if(cam.zoom > camMax.trsZ)
			cam.zoom = camMax.trsZ;
		if(cam.zoom < camMin.trsZ)
			cam.zoom = camMin.trsZ;
		
		refresh();
		drawable.display();
		*/
		
		/* smooth */
		float destZoom = cam.zoom;  
	    if (wheelRotation < 0) //Roll up (zoom in)
			destZoom /= zoomStep;
		if (wheelRotation > 0) //Roll down (zoom out)
			destZoom *= zoomStep;
		
		if(destZoom > camMax.zoom)
			destZoom = camMax.zoom;
		if(destZoom < camMin.zoom)
			destZoom = camMin.zoom;
		
		float steps = 6.0f;
		float dz = (destZoom - cam.zoom) / steps;
		
		for(int i=0; i < steps; i++){
			cam.zoom += dz;
			refreshViewToWinRatio();
			drawable.display();
		}
	}

	public void movePlayer(int x, int y) {
//		float px =  x * viewToWinRatio - 100.0f;
		float px =  x * 0.395f - 100.0f;
//		float py = -y * viewToWinRatio + 40.0f;
		float py = -y * 0.4f + 40.0f;
		//controledPlayer.moveJoint(px, py);
		
		if(!MasterListener.outList.contains("move_joint")) {
			try {
				MasterListener.out.writeObject("move_joint");
				MasterListener.out.writeFloat(px);
				MasterListener.out.writeFloat(py);
				
				MasterListener.outList.add("move_joint");
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
//		System.err.println(px + "; " + py + "; ratio= " + viewToWinRatio);
	}
	
	
	public void homeView() {
		gotoView(home);
	}
	
	public void frontView() {
		gotoView(front);
	}

	public void topView() {
		gotoView(top);		
	}

	public void rightView() {
		gotoView(right);
	}

	private void gotoView(Location dest){
		float steps = 150.0f;
		
		// Try to use the shortest path to home
		// (don't go more the 180 degrees)
		if (cam.rotZ > dest.rotZ + 180.0f ) {
			// percent of steps that need to be done this way
			float stepsRatio = (360.0f - cam.rotZ) / (360.0f - cam.rotZ + dest.rotZ);
			float stepsFirstStage = steps * stepsRatio;
			float stepsSecondStage = steps * (1.0f - stepsRatio);
			System.err.println("stepsRatio = " + stepsRatio +
					"     fistStage = " + stepsFirstStage +
					"     secondStage = " + stepsSecondStage);		
			
			// then we first would prefer to go to maxRotZ
			// and then go home
			float trsXd = (cam.trsX - dest.trsX) / (stepsFirstStage);
			float trsYd = (cam.trsY - dest.trsY) / (stepsFirstStage);
			float trsZd = (cam.trsZ - dest.trsZ) / (stepsFirstStage);
			
			float zoomd = (cam.zoom - dest.zoom) / (stepsFirstStage);
			float rotXd = (cam.rotX - dest.rotX) / (stepsFirstStage);
			float rotZd = (cam.rotZ - 360.0f) / (stepsFirstStage);
			
			for (int i = 0; i < (stepsFirstStage); i++) {
				cam.trsX -= trsXd * stepsRatio;
				cam.trsY -= trsYd * stepsRatio;
				cam.trsZ -= trsZd * stepsRatio;
				cam.zoom -= zoomd * stepsRatio;
				cam.rotX -= rotXd * stepsRatio;
				cam.rotZ -= rotZd;
				
				drawable.display();
				
			}
			cam.rotZ = 0.0f;
			steps = stepsSecondStage;			
		}
			
		float trsXd = (cam.trsX - dest.trsX) / steps;
		float trsYd = (cam.trsY - dest.trsY) / steps;
		float trsZd = (cam.trsZ - dest.trsZ) / steps;
		
		float zoomd = (cam.zoom - dest.zoom) / steps;
		float rotXd = (cam.rotX - dest.rotX) / steps;
		//float rotYd = (rotY - dest.rotY) / steps;
		float rotZd = (cam.rotZ - dest.rotZ) / steps;
		
		for (int i = 0; i < steps; i++) {
			cam.trsX -= trsXd;
			cam.trsY -= trsYd;
			cam.trsZ -= trsZd;
			cam.zoom -= zoomd;
			cam.rotX -= rotXd;
			cam.rotZ -= rotZd;
			
			
			
			drawable.display();		
		}
		
		cam.trsX = dest.trsX;
		cam.trsY = dest.trsY;
		cam.trsZ = dest.trsZ;
		cam.zoom = dest.zoom;
		cam.rotX = dest.rotX;		
		cam.rotZ = dest.rotZ;	
		refreshViewToWinRatio();
		
	}
	
	public void toggleWireframe() {
		
		wireframe = !wireframe; 
		System.err.println("wireframe = " + wireframe);
		drawable.display();
	}	
	
	public void toggleShading() {
		
		flatShading = !flatShading;				
		System.err.println("flatSading = " + flatShading);
		drawable.display();
	}

	public void toggleCullFaces() {
		
		cullFaces = !cullFaces;
		System.err.println("cullFaces = " + cullFaces);
		drawable.display();
	}

	public void toggleAverage() {			
//		MasterListener.shapesBag.average = !MasterListener.shapesBag.average; 	
//		System.err.println("JetShadow.shapesBag.average = " + MasterListener.shapesBag.average);
//		drawable.display();
	}

	@Override
	public void init(GLAutoDrawable drawable) {
		// Use debug pipeline
		// drawable.setGL(new DebugGL(drawable.getGL()));
	
		GL gl = drawable.getGL();
		System.err.println("INIT GL IS: " + gl.getClass().getName());
		
		this.drawable = drawable;
		
		// Enable VSync
		gl.setSwapInterval(1);
	
//	    gl.glFrontFace(GL.GL_CCW);
	    gl.glEnable(GL.GL_CULL_FACE);		
	    gl.glEnable(GL.GL_DEPTH_TEST);
		gl.glDepthFunc(GL.GL_LEQUAL);
		gl.glHint(GL.GL_PERSPECTIVE_CORRECTION_HINT, GL.GL_NICEST);
		
		gl.glEnable(GL.GL_BLEND);
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
//		gl.glEnable(GL.GL_LINE_SMOOTH);
//		gl.glEnable(GL.GL_POINT_SMOOTH);
//		gl.glEnable(GL.GL_POLYGON_SMOOTH);
		
		
		// Setup the drawing area and shading mode
	    gl.glClearColor(70/255.0f,  84/255.0f, 100/255.0f, 1.0f);
//		gl.glShadeModel(GL.GL_FLAT);
//		gl.glShadeModel(GL.GL_SMOOTH);
		
		
		cam.setTo(home);
	}

	@Override
		public void display(GLAutoDrawable drawable) {
			GL gl = drawable.getGL();
	
			// Clear the drawing area
			gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
			
			// Reset the current matrix to the "identity"
			gl.glLoadIdentity();
	
		    //{ Viewer =================================================
	//	    gl.glPushMatrix();
			
			// Rotate around (dynamic) local center of the camera			
			gl.glTranslatef(0.0f, 0.0f, cam.zoom);
			gl.glRotatef(cam.rotX, 1.0f, 0.0f, 0.0f);
		    gl.glRotatef(cam.rotZ, 0.0f, 0.0f, 1.0f);		    
//		    arrowsToViewer.displayIfVisible(drawable);
		    
		    gl.glTranslatef(cam.trsX, cam.trsY, cam.trsZ);
		    
		    
			if (flatShading){
				gl.glShadeModel(GL.GL_FLAT);			
			} else {
				gl.glShadeModel(GL.GL_SMOOTH);
			}		
			
			if (cullFaces){			
				gl.glEnable(GL.GL_CULL_FACE);
			} else {
				gl.glDisable(GL.GL_CULL_FACE);
			}
			
		    if(wireframe)
		        //gl.glPolygonMode(GL.GL_BACK,GL.GL_LINE);
		    	gl.glPolygonMode(GL.GL_FRONT_AND_BACK,GL.GL_LINE);
		    else
		        gl.glPolygonMode(GL.GL_FRONT_AND_BACK,GL.GL_FILL);
			
		    //{ Main drawing ===========================================
		    // rest of the drawing will be done with different EventListeners
			//} Main drawing ===========================================	
		    
	//		gl.glPopMatrix();
			//} Viewer =================================================
	
			// Flush all drawing operations to the graphics card
	//		gl.glFlush();
		}

	@Override
	public void reshape(GLAutoDrawable drawable, int x, int y, int width,
			int height) {
		GL gl = drawable.getGL();
		GLU glu = new GLU();
		
		winHeight = height;
		winWidth  = width;
		refreshViewToWinRatio();

		gl.glViewport(0, 0, winWidth, winHeight);
		gl.glMatrixMode(GL.GL_PROJECTION);
		gl.glLoadIdentity();
		glu.gluPerspective(fovy, winAspect, nearClip, farClip);
		gl.glMatrixMode(GL.GL_MODELVIEW);
		gl.glLoadIdentity();
	}

	@Override
	public void displayChanged(GLAutoDrawable drawable, boolean modeChanged,
			boolean deviceChanged) {
	}

	
}


/**
 * Wrapper for Vector3f from vecmath.
 * @author TomDalek
 */
class MyVector3f extends Vector3f {
	private static final long serialVersionUID = 1L;

	
	/**
	 * I'm using this as a glue for the JOGL in places like:<br>
	 *  gl.glNormal3fv(normal.get(), 0);
	 *  
	 * @return This vector as an array of floats.
	 */
	public float[] get(){
		float[] array = new float[3];		
		this.get(array);
		return array;
	}
	
	/** 
	 * I'm using this for Viewer, because camera is stored in polar coordinations
	 * @param rotZ - cameras rotation along local Z in degrees
	 * @param rotX - cameras rotation along local X in degrees
	 * @param length - mouse movement displacement
	 */
	public void setFromPolor(float rotZ, float rotX, float length){	
		
		// My way for global Z-up
		// rotX=0 is Azymuth, rotX=-90 is ground.
		// rotZ=0 is Y-	    , rotZ=90 is X- (CW rotation).
		float sinRotZ = (float)Math.sin(Math.toRadians(-rotZ-90.0f));
		float cosRotZ = (float)Math.cos(Math.toRadians(-rotZ-90.0f));
		float sinRotX = (float)Math.sin(Math.toRadians(-rotX));
		float cosRotX = (float)Math.cos(Math.toRadians(-rotX));
				
		this.x = length * sinRotX * cosRotZ;
		this.y = length * sinRotX * sinRotZ;
		this.z = length * cosRotX;
	}
}